# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command that statically validates gcloud commands for corectness.

To validate a command, run:

```
gcloud meta lint-gcloud-commands --command-string="gcloud compute instances
list"
```

To validate a list of commands in a file:

1) Create a JSON file with the following format:

```
[
  {
    "command_string": "gcloud compute instances list",
  },
  {
    "command_string": "gcloud compute instances describe my-instance",
  }
]
```

2) Then run the command:

```
gcloud meta lint-gcloud-commands --commands-file=commands.json
```

Commands can also be associated with an ID, which will be used to identify the
command in the output. Simply run:

```
gcloud meta lint-gcloud-commands --commands-file=commands.json --serialize
```
This will associated each command with using the index it was found in the file
as the ID. If you want to associate a command with a specific ID, you can do so
by adding the `id` field to the command in the JSON file. For example:

```
[
  {
    "command_string": "gcloud compute instances list",
    "id": 0,
  },
  {
    "command_string": "gcloud compute instances describe my-instance",
    "id": 1,
  }
]
```

This will output the validation results in the following format:

```
{"0": [{<OUTPUT_1>}], "1": [{<OUTPUT_2>}]}
"""


import argparse
import copy
import json
import os
import re
import shlex
from typing import collections

from googlecloudsdk import gcloud_main
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as gcloud_exceptions
from googlecloudsdk.command_lib.meta import generate_argument_spec
from googlecloudsdk.core import log
from googlecloudsdk.core import yaml
from googlecloudsdk.core.util import files


_PARSING_OUTPUT_TEMPLATE = {
    'command_string': None,
    'success': False,
    'command_args': None,
    'command_string_no_args': None,
    'args_structure': {},
    'error_message': None,
    'error_type': None,
}

_IGNORE_ARGS = ['--help']


class CommandValidationError(Exception):
  pass


def _read_commands_from_file(commands_file):
  """Reads commands from a JSON file."""
  with files.FileReader(commands_file) as f:
    command_file_data = json.load(f)
  ref_id = 0
  command_strings = {}
  needs_id = any(command_data.get('id') for command_data in command_file_data)
  for idx, command_data in enumerate(command_file_data):
    command_id = command_data.get('id')
    if needs_id and command_id is None:
      raise ValueError(
          'Not all commands have an ID. Id for command'
          f' {command_data["command_string"]} is None.'
      )
    command_strings[command_data['command_string']] = (
        command_id or (ref_id + idx))
  return command_strings


def _separate_command_arguments(command_string: str):
  """Move all flag arguments to back."""
  command_string = command_string.split('#')[0]
  try:
    # Split arguments
    if os.name == 'nt':
      command_arguments = shlex.split(command_string, posix=False)
    else:
      command_arguments = shlex.split(command_string)
  except Exception:  # pylint: disable=broad-except
    raise CommandValidationError(
        'Command could not be validated due to unforeseen edge case.'
    ) from None
  # Move any flag arguments to end of command.
  flag_args = [arg for arg in command_arguments if arg.startswith('--')]
  command_args = [arg for arg in command_arguments if not arg.startswith('--')]
  return command_args + flag_args


def _add_equals_to_flags(command):
  """Adds equals signs to gcloud command flags, except for format and help flags."""

  pattern = (  # Matches flag name and its value (excluding format and help)
      r'(--[a-zA-Z0-9-]+) +([^-][^ ]*)'
  )
  replacement = r'\1=\2'  # Inserts equals sign between flag and
  modified_command = re.sub(pattern, replacement, command)
  # Remove = from flags without explicit values
  modified_command = re.sub(r'(--[a-zA-Z0-9-]+)= ', r'\1 ', modified_command)
  return modified_command


def formalize_gcloud_command(command_str):
  command_str = _add_equals_to_flags(command_str)
  command_str = (
      command_str.replace('--project=PROJECT ', '--project=my-project ')
      .replace('--project=PROJECT_ID ', '--project=my-project ')
      .replace('$PROJECT_ID ', 'my-project ')
      .replace('YOUR_PROJECT_ID ', 'my-project ')
  )
  return command_str


def _extract_gcloud_commands(text):
  """Extracts code snippets from fenced code blocks within a text string.

  Args:
      text: The text string containing fenced code blocks.

  Returns:
      A list of extracted code snippets.
  """
  text = bytes(text, 'utf-8').decode('unicode_escape')
  fenced_pattern = r'```(?:[\w ]+\n)?(.*?)```'
  indented_pattern = (  # 3-8 indented spaces as arbitray nums
      r'(?: {3-8}|\t)(.*?)(?:\n\S|\n$)'
  )
  combined_pattern = re.compile(
      f'{fenced_pattern}|{indented_pattern}', re.DOTALL
  )

  code_snippets = []
  for match in combined_pattern.finditer(
      text
  ):  # use finditer instead of findall
    command_str = match.group(1).strip()
    if 'gcloud ' not in command_str or not command_str.startswith('gcloud'):
      continue
    for cmd in command_str.split('gcloud '):
      cmd_new_lines = cmd.split('\n')
      if len(cmd_new_lines) >= 1 and cmd_new_lines[0].strip():
        command_str = formalize_gcloud_command(cmd_new_lines[0].strip())
        code_snippets.append(f'gcloud {command_str}')
  return code_snippets


def _get_command_node(command_arguments):
  """Returns the command node for the given command arguments."""
  cli = gcloud_main.CreateCLI([])
  command_arguments = command_arguments[1:]
  current_command_node = cli._TopElement()  # pylint: disable=protected-access
  for argument in command_arguments:
    if argument.startswith('--'):
      break
    child_command_node = current_command_node.LoadSubElement(argument)
    if not child_command_node:
      break
    current_command_node = child_command_node
  return current_command_node


def _get_command_no_args(command_node):
  """Returns the command string without any arguments."""
  return ' '.join(command_node.ai.command_name)


def _get_command_args_tree(command_node):
  """Returns the command string without any arguments."""
  argument_tree = generate_argument_spec.GenerateArgumentSpecifications(
      command_node
  )
  return argument_tree


def _get_positional_metavars(args_tree):
  """Returns a dict of positional metavars."""

  positional_args = []

  def _process_arg(node):
    if 'name' in node and node.get('positional', False) and node['name']:
      positional_args.append(node['name'])

  def _traverse_arg_group(node):
    for arg in node:
      _traverse_tree(arg)

  def _traverse_tree(node):
    if 'group' in node:
      group = node['group']['arguments']
      _traverse_arg_group(group)
    else:
      _process_arg(node)

  for node in args_tree:
    _traverse_tree(node)
  return positional_args


def _normalize_command_args(command_args, args_tree):
  """Normalizes command args for storage."""

  positionals_used = set()
  arg_name_value = {}
  positional_args_in_tree = _get_positional_metavars(args_tree['arguments'])

  def _sort_command_args(args):
    """Sorts command arguments.

    Arguments starting with '--' are placed at the back, and all arguments are
    ordered alphabetically.

    Args:
      args: The command arguments to sort.

    Returns:
      The sorted command arguments.
    """
    flag_args = sorted([arg for arg in args if arg.startswith('--')])
    positional_args = [arg for arg in args if not arg.startswith('--')]
    return positional_args + flag_args

  command_args = _sort_command_args(command_args)

  def _get_next_available_positional_arg():
    for positional_metavar in positional_args_in_tree:
      if positional_metavar not in positionals_used:
        command_value = command_arg
        command_arg_name = positional_metavar.upper()
        positionals_used.add(positional_metavar)
        return command_arg_name, command_value
    return None, None

  for arg_index, command_arg in enumerate(command_args):
    command_arg_name = command_arg
    if command_arg.startswith('--'):
      equals_index = command_arg.find('=')
      if equals_index != -1:
        command_arg_name = command_arg[:equals_index]
        command_value = command_arg[equals_index + 1 :]
      else:
        command_value = ''
    else:
      # Positional argument
      command_arg_name, command_value = _get_next_available_positional_arg()
      # Arg should be included in output, regardless of whether it a real
      # positional arg or not.
      command_arg_name = command_arg_name or command_arg
      command_value = command_value or ''
    arg_name_value[command_arg_name] = {
        'value': command_value,
        'index': arg_index,
    }
  return collections.OrderedDict(
      sorted(arg_name_value.items(), key=lambda item: item[1]['index'])
  )


@base.UniverseCompatible
class GenerateCommand(base.Command):
  """Generate YAML file to implement given command.

  The command YAML file is generated in the --output-dir directory.
  """

  _INDEXED_VALIDATION_RESULTS = collections.OrderedDict()
  _SERIALIZED_VALIDATION_RESULTS = collections.OrderedDict()
  _VALIDATION_RESULTS = []
  index_results = False
  serialize_results = False

  def _validate_command(self, command_string, ref_id=0):
    """Validate a single command."""
    command_string = formalize_gcloud_command(command_string)
    command_arguments = _separate_command_arguments(command_string)
    command_success, command_node, flag_arguments = (
        self._validate_command_prefix(command_arguments, command_string, ref_id)
    )
    if not command_success:
      return
    flag_success = self._validate_command_suffix(
        command_node, flag_arguments, command_string, ref_id
    )
    if not flag_success:
      return
    self._store_validation_results(True, command_string, ref_id, flag_arguments)

  def _validate_commands_from_file(self, commands_file):
    """Validate multiple commands given in a file."""
    commands = _read_commands_from_file(commands_file)
    for command, ref_id in commands.items():
      try:
        self._validate_command(command, ref_id)
      except Exception as e:  # pylint: disable=broad-except
        self._store_validation_results(
            False,
            command,
            ref_id,
            None,
            f'Command could not be validated: {e}',
            'CommandValidationError',
        )

  def _validate_commands_from_text(self, commands_text_file):
    """Validate multiple commands given in a text string."""
    with files.FileReader(commands_text_file) as f:
      text = f.read()
    commands = _extract_gcloud_commands(text)
    for ref_id, command in enumerate(commands):
      self._validate_command(command, ref_id)

  def _validate_command_prefix(self, command_arguments, command_string, ref_id):
    """Validate that the argument string contains a valid command or group."""
    cli = gcloud_main.CreateCLI([])
    # Remove "gcloud" from command arguments.
    command_arguments = command_arguments[1:]
    index = 0
    current_command_node = cli._TopElement()  # pylint: disable=protected-access
    for argument in command_arguments:
      # If this hits, we've found a command group with a flag passed.
      # e.g. gcloud compute --help
      if argument.startswith('--'):
        return (
            True,
            current_command_node,
            command_arguments[index:],
        )
      # Attempt to load next section of command path.
      current_command_node = current_command_node.LoadSubElement(argument)
      # If not a valid section of command path, fail validation.
      if not current_command_node:
        self._store_validation_results(
            False,
            command_string,
            ref_id,
            command_arguments[index:],
            f"Invalid choice: '{argument}'",
            'UnrecognizedCommandError',
        )
        return False, None, None
      index += 1
      # If command path is valid and is a command, return the command node.
      if not current_command_node.is_group:
        return (
            True,
            current_command_node,
            command_arguments[index:],
        )

    # If we make it here then only a command group has been provided.
    remaining_flags = command_arguments[index:]
    if not remaining_flags:
      self._store_validation_results(
          False,
          command_string,
          ref_id,
          command_arguments[index:],
          'Command name argument expected',
          'UnrecognizedCommandError',
      )
      return False, None, None
    # If we've iterated through the entire list and end up here, something
    # unpredicted has happened.
    raise CommandValidationError(
        'Command could not be validated due to unforeseen edge case.'
    )

  def _validate_command_suffix(
      self, command_node, command_arguments, command_string, ref_id
  ):
    """Validates that the given flags can be parsed by the argparse parser."""
    for ignored_arg in _IGNORE_ARGS:
      if ignored_arg in command_arguments:
        return True
    found_parent = False
    if command_arguments:
      for command_arg in command_arguments:
        if (
            '--project' in command_arg
            or '--folder' in command_arg
            or '--organization' in command_arg
        ):
          found_parent = True
    if not command_arguments:
      command_arguments = []
    if not found_parent:
      command_arguments.append('--project=myproject')
    try:
      command_node._parser.parse_args(command_arguments, raise_error=True)  # pylint: disable=protected-access
    except (
        files.MissingFileError,
        gcloud_exceptions.BadFileException,
        yaml.FileLoadError,
    ):
      pass
    except argparse.ArgumentError as e:
      if 'No such file or directory' in str(e):
        return True
      self._store_validation_results(
          False,
          command_string,
          ref_id,
          command_arguments,
          str(e),
          type(e).__name__,
      )
      return False
    return True

  def _store_validation_results(
      self,
      success,
      command_string,
      ref_id,
      command_args=None,
      error_message=None,
      error_type=None,
  ):
    """Store information related to the command validation."""
    validation_output = copy.deepcopy(_PARSING_OUTPUT_TEMPLATE)
    validation_output['command_string'] = command_string
    try:
      command_node = _get_command_node(
          _separate_command_arguments(command_string)
      )
      validation_output['command_string_no_args'] = _get_command_no_args(
          command_node
      )
      validation_output['args_structure'] = _get_command_args_tree(command_node)
    except Exception:  # pylint: disable=broad-except
      validation_output['command_string_no_args'] = command_string
    if command_args:
      validation_output['command_args'] = _normalize_command_args(
          command_args, validation_output['args_structure']
      )
    validation_output['success'] = success
    validation_output['error_message'] = error_message
    validation_output['error_type'] = error_type
    sorted_validation_output = collections.OrderedDict(
        sorted(validation_output.items())
    )
    if self.serialize_results:
      if ref_id not in self._SERIALIZED_VALIDATION_RESULTS:
        self._SERIALIZED_VALIDATION_RESULTS[ref_id] = [sorted_validation_output]
      else:
        self._SERIALIZED_VALIDATION_RESULTS[ref_id].append(
            sorted_validation_output
        )
    if self.index_results:
      self._INDEXED_VALIDATION_RESULTS[command_string] = (
          sorted_validation_output
      )
    else:
      self._VALIDATION_RESULTS.append(sorted_validation_output)

  def _log_validation_results(self):
    """Output collected validation results."""
    if self.index_results:
      log.out.Print(json.dumps(self._INDEXED_VALIDATION_RESULTS))
    elif self.serialize_results:
      log.out.Print(json.dumps(self._SERIALIZED_VALIDATION_RESULTS))
    else:
      log.out.Print(json.dumps(self._VALIDATION_RESULTS))

  @staticmethod
  def Args(parser):
    command_group = parser.add_group(mutex=True)
    command_group.add_argument(
        '--command-string',
        help='Gcloud command to statically validate.',
    )
    command_group.add_argument(
        '--commands-file',
        help='JSON file containing list of gcloud commands to validate.',
    )
    command_group.add_argument(
        '--commands-text-file',
        help=(
            'Raw text containing gcloud command(s) to validate. For example,'
            ' the commands could be in fenced code blocks or indented code'
            ' blocks.'
        ),
    )
    parser.add_argument(
        '--serialize',
        action='store_true',
        help='Output results in a dictionary serialized by reference id.',
    )

  def Run(self, args):
    if args.serialize:
      self.serialize_results = True
    if args.IsSpecified('command_string'):
      self._validate_command(args.command_string)
    elif args.IsSpecified('commands_text_file'):
      self._validate_commands_from_text(args.commands_text_file)
    else:
      self._validate_commands_from_file(args.commands_file)
    self._log_validation_results()



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Updates a Backup and DR Backup."""

import argparse
from apitools.base.py import exceptions as apitools_exceptions
from googlecloudsdk.api_lib.backupdr import util
from googlecloudsdk.api_lib.backupdr.backups import BackupsClient
from googlecloudsdk.api_lib.util import exceptions
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as calliope_exceptions
from googlecloudsdk.command_lib.backupdr import flags
from googlecloudsdk.command_lib.backupdr import util as command_util
from googlecloudsdk.core import log


def _add_common_args(parser: argparse.ArgumentParser) -> None:
  """Specifies additional command flags.

  Args:
    parser: argparse.Parser: Parser object for command line inputs.
  """
  base.ASYNC_FLAG.AddToParser(parser)
  base.ASYNC_FLAG.SetDefault(parser, True)
  flags.AddBackupResourceArg(
      parser,
      'Name of the backup to update.',
  )
  flags.AddUpdateBackupFlags(parser)


def _get_update_mask(args) -> str:
  updated_fields = []
  if args.IsSpecified('enforced_retention_end_time'):
    updated_fields.append('enforcedRetentionEndTime')
  if args.IsSpecified('expire_time'):
    updated_fields.append('expireTime')
  label_flags = ['update_labels', 'remove_labels', 'clear_labels']
  if any(args.IsSpecified(flag) for flag in label_flags):
    updated_fields.append('labels')
  return ','.join(updated_fields)


@base.DefaultUniverseOnly
@base.ReleaseTracks(base.ReleaseTrack.GA)
class Update(base.UpdateCommand):
  """Update the specified Backup."""

  detailed_help = {
      'BRIEF': 'Updates a specific backup',
      'DESCRIPTION': '{description}',
      'EXAMPLES': """\
        To update the enforced retention of a backup sample-backup in backup vault sample-vault, data source
        sample-ds, project sample-project and location us-central1, run:

          $ {command} sample-backup --backup-vault=sample-vault --data-source=sample-ds --project=sample-project --location=us-central1 --enforced-retention-end-time="2025-02-14T01:10:20Z"

        To update the labels on a backup named `my-backup` in the `test-bv` vault and `us-central1` location, run:

            $ {command} my-backup --backup-vault=test-bv --location=us-central1 --update-labels=env=prod,team=storage

        To remove the `env` label from the same backup, run:

            $ {command} my-backup --backup-vault=test-bv --location=us-central1 --remove-labels=env

        To clear all labels from the backup, run:

            $ {command} my-backup --backup-vault=test-bv --location=us-central1 --clear-labels
        """,
  }

  @staticmethod
  def Args(parser: argparse.ArgumentParser) -> None:
    _add_common_args(parser)

  def ParseUpdate(self, backup, args, client):
    updated_enforced_retention = command_util.VerifyDateInFuture(
        args.enforced_retention_end_time, 'enforced-retention-end-time'
    )

    expire_time = command_util.VerifyDateInFuture(
        args.expire_time, 'expire-time'
    )

    return client.ParseUpdate(
        backup,
        updated_enforced_retention,
        expire_time,
        args.update_labels,
        args.remove_labels,
        args.clear_labels,
    )

  def Run(self, args):
    """Constructs and sends request.

    Args:
      args: argparse.Namespace, An object that contains the values for the
        arguments specified in the .Args() method.

    Returns:
      ProcessHttpResponse of the request made.
    """
    api_version = util.GetApiVersion(self.ReleaseTrack())
    client = BackupsClient(api_version=api_version)
    is_async = args.async_
    backup = args.CONCEPTS.backup.Parse()

    label_flags = ['update_labels', 'remove_labels', 'clear_labels']
    has_label_flag = any(args.IsSpecified(flag) for flag in label_flags)
    has_enforced_retention = args.IsSpecified('enforced_retention_end_time')
    has_expire_time = args.IsSpecified('expire_time')
    has_time_flag = has_enforced_retention or has_expire_time

    if not (has_label_flag or has_time_flag):
      raise calliope_exceptions.RequiredArgumentException(
          '--update-labels | --remove-labels | --clear-labels | '
          '(--enforced-retention-end-time | --expire-time)',
          'At least one of the update flags must be provided.',
      )

    try:
      parsed_backup = self.ParseUpdate(backup, args, client)

      update_mask = _get_update_mask(args)

      operation = client.Update(
          backup,
          parsed_backup,
          update_mask=update_mask,
      )
    except apitools_exceptions.HttpError as e:
      raise exceptions.HttpException(e, util.HTTP_ERROR_FORMAT)

    if is_async:
      log.UpdatedResource(
          backup.RelativeName(),
          kind='backup',
          is_async=True,
          details=util.ASYNC_OPERATION_MESSAGE.format(operation.name),
      )
      return operation

    response = client.WaitForOperation(
        operation_ref=client.GetOperationRef(operation),
        message=(
            'Updating backup [{}]. (This operation usually takes less than 1'
            ' minute.)'.format(backup.RelativeName())
        ),
        has_result=False,
    )
    log.UpdatedResource(backup.RelativeName(), kind='backup')

    return response


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class UpdateAlpha(Update):
  """Update the specified Backup."""

  @staticmethod
  def Args(parser: argparse.ArgumentParser) -> None:
    _add_common_args(parser)

  def ParseUpdate(self, backup, args, client):
    updated_enforced_retention = command_util.VerifyDateInFuture(
        args.enforced_retention_end_time, 'enforced-retention-end-time'
    )

    expire_time = command_util.VerifyDateInFuture(
        args.expire_time, 'expire-time'
    )

    return client.ParseUpdate(
        backup,
        updated_enforced_retention,
        expire_time,
        args.update_labels,
        args.remove_labels,
        args.clear_labels,
    )

  def GetUpdateMask(self, args):
    return _get_update_mask(args)


@base.ReleaseTracks(base.ReleaseTrack.BETA)
class UpdateBeta(Update):
  """Update the specified Backup."""

  @staticmethod
  def Args(parser: argparse.ArgumentParser) -> None:
    _add_common_args(parser)

  def ParseUpdate(self, backup, args, client):
    updated_enforced_retention = command_util.VerifyDateInFuture(
        args.enforced_retention_end_time, 'enforced-retention-end-time'
    )

    expire_time = command_util.VerifyDateInFuture(
        args.expire_time, 'expire-time'
    )

    return client.ParseUpdate(
        backup,
        updated_enforced_retention,
        expire_time,
        args.update_labels,
        args.remove_labels,
        args.clear_labels,
    )

  def GetUpdateMask(self, args):
    return _get_update_mask(args)



# -*- coding: utf-8 -*- #
# Copyright 2026 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to submit a Device Run instrumentation session."""

import datetime
import os
import uuid

from googlecloudsdk.api_lib import device_run
from googlecloudsdk.api_lib.storage import storage_api
from googlecloudsdk.api_lib.storage import storage_util
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as calliope_exceptions
from googlecloudsdk.command_lib.device_run import resource_args
from googlecloudsdk.core import log
from googlecloudsdk.core import properties


def _GetRunId():
  timestamp = datetime.datetime.now().strftime('%Y-%m-%d_%H:%M.%S%f')
  random_suffix = uuid.uuid4().hex[:4].upper()
  return f'{timestamp}_{random_suffix}'


def _UploadFileIfNeeded(path, bucket_name, storage_client, run_id):
  """Uploads a local file to GCS if it is not already a GCS path."""
  if path.startswith('gs://'):
    return path
  log.status.Print(f'Uploading [{path}].')
  target_gcs_path = (
      f'gs://{bucket_name}/automation/inputs/{run_id}/{os.path.basename(path)}'
  )
  target_obj_ref = storage_util.ObjectReference.FromUrl(target_gcs_path)
  storage_client.CopyFileToGCS(path, target_obj_ref)
  return target_gcs_path


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class Instrumentation(base.Command):
  """Submit a Device Run session with an instrumentation job."""

  @staticmethod
  def Args(parser):
    resource_args.AddLocationResourceArg(parser, 'submit session')
    parser.add_argument(
        '--device',
        required=True,
        type=str,
        help='Id of the device to run the test on.',
    )
    parser.add_argument(
        '--test',
        required=True,
        type=str,
        help=(
            'The path to the binary file containing instrumentation tests.'
            ' Supports both Google Cloud Storage (`gs://...`) paths and local'
            ' filesystem paths. Any local file will be uploaded to Google Cloud'
            ' Storage prior to test execution.'
        ),
    )
    parser.add_argument(
        '--apps',
        metavar='SOURCE_PATH',
        type=arg_parsers.ArgList(),
        default=[],
        help=(
            'A list of application binary files to install before running the'
            ' test. The order of the applications in the list determines their'
            ' installation order on the device. Supports both Google Cloud'
            ' Storage (`gs://...`) paths and local filesystem paths. Any local'
            ' file will be uploaded to Google Cloud Storage prior to test'
            ' execution.'
        ),
    )
    parser.add_argument(
        '--bucket-name',
        type=str,
        help="""\
The name of a Google Cloud Storage bucket to store test artifacts, including
local input files, test output files, and smart sharding timing records. If
not specified, a default bucket named `<project>-devicerun` will be used or
created.

The Google Cloud Storage bucket layout will be structured as follows:

  * gs://{bucket-name}/automation/
    * inputs/
      * 2026-05-29_10:13.026220_JZTM/
        * app-debug.apk
        * app-debug-androidTest.apk
    * sessions/
      * my-session-id1/
        * job-000/
          * execution-000/
            * junit.xml
            * ...
          * execution-001/
            * junit.xml
            * ...
          * merged_junit.xml
      * my-session-id2/
        * ...
    * smart-sharding/
      * sharding-record-A.yaml
      * sharding-record-B.yaml
""",
    )
    parser.add_argument(
        '--test-runner-class',
        type=str,
        help=(
            'The instrumentation test runner class to use. If not specified, a'
            ' default runner class will be determined by examining the'
            ' application\'s manifest.'
        ),
    )
    parser.add_argument(
        '--test-targets',
        metavar='KEY=VALUE',
        type=arg_parsers.ArgDict(operators={'=': None, ' ': None}),
        action=arg_parsers.UpdateActionWithAppend,
        default={},
        help="""\
A list of test targets or target filters to run. Each target must be fully
qualified with the package name or class name, in one of these formats:

  * `package package_name`
  * `notPackage com.package.to.skip`
  * `class package_name.class_name`
  * `class package_name.class_name#method_name`
  * `notClass com.foo.ClassToSkip`
  * `notClass com.foo.ClassName#testMethodToSkip`
  * `annotation com.foo.AnnotationToRun`
  * `notAnnotation com.foo.AnnotationToSkip`
  * `size [small|medium|large]`

Formats like `testfile` or `notTestfile` won't be supported. If empty, all
targets in the module will be run.
""",
    )
    parser.add_argument(
        '--additional-test-options',
        metavar='KEY=VALUE',
        type=arg_parsers.ArgDict(operators={'=': None, ' ': None}),
        action=arg_parsers.UpdateAction,
        default={},
        help=(
            'A dictionary of additional options to pass to the instrumentation'
            ' test. Formats supported in test_targets are not allowed to be'
            ' used here.'
        ),
    )
    parser.add_argument(
        '--other-files-to-push',
        metavar='SOURCE=DEST',
        type=arg_parsers.ArgDict(operators={'=': None, ' ': None}),
        action=arg_parsers.UpdateAction,
        default={},
        help=(
            'A dictionary of additional files to be pushed to the device before'
            ' running the test. The key is the source path of the file'
            ' (supports both Google Cloud Storage and local paths; local files'
            ' will be uploaded to Google Cloud Storage prior to test'
            ' execution), and the value is the destination path on the device.'
        ),
    )
    parser.add_argument(
        '--paths-to-pull',
        metavar='DEVICE_PATH',
        type=arg_parsers.ArgList(),
        default=[],
        help=(
            'A list of file or directory paths to pull from the device'
            ' following test completion.'
        ),
    )
    parser.add_argument(
        '--video',
        type=str,
        choices=['always', 'on-failure'],
        help=(
            'Specifies when to record video of the device screen during the'
            ' test run. Accepted values are `always` or `on-failure`.'
        ),
    )
    parser.add_argument(
        '--labels',
        metavar='KEY=VALUE',
        type=arg_parsers.ArgDict(operators={'=': None, ' ': None}),
        action=arg_parsers.UpdateAction,
        default={},
        help=(
            'A dictionary of user-defined key-value labels to attach to the'
            ' session.'
        ),
    )
    parser.add_argument(
        '--sharding-option',
        type=str,
        choices=['uniform', 'smart'],
        help=(
            'Specifies the sharding strategy to partition the test execution'
            ' across multiple device allocations. Accepted values are `uniform`'
            ' or `smart`.'
        ),
    )
    parser.add_argument(
        '--uniform-shard-count',
        type=int,
        help=(
            'Specifies the total number of shards to create for uniform'
            ' sharding. Required when `--sharding-option=uniform`.'
        ),
    )
    parser.add_argument(
        '--target-smart-shard-time',
        type=int,
        help=(
            'Specifies the targeted execution time (in seconds) per shard for'
            ' smart sharding. Required when `--sharding-option=smart`.'
        ),
    )
    parser.add_argument(
        '--smart-sharding-record-name',
        type=str,
        help=(
            'Specifies the name of the smart sharding record file, excluding'
            ' the file extension. Required when `--sharding-option=smart`. This'
            ' file is located in the Google Cloud Storage bucket specified by'
            ' `--bucket-name` under the `smart-sharding/` directory with a'
            ' `.yaml` extension. If the file does not exist, it will be created'
            ' automatically; otherwise, its contents will be updated upon'
            ' session completion.'
        ),
    )
    parser.add_argument(
        '--max-attempts',
        type=int,
        help=(
            'Specifies the maximum number of execution attempts per test shard'
            ' to handle flakiness or infrastructure errors. If not specified,'
            ' defaults to 1.'
        ),
    )
    parser.add_argument(
        '--instrumentation-timeout',
        type=arg_parsers.Duration(lower_bound='1m', upper_bound='1h'),
        help=(
            'Specifies the maximum duration allowed for the instrumentation'
            ' test run (e.g., `10m`, `20s`, `1h`). The valid range is `1m` to'
            ' `1h`. If not specified, defaults to `5m`.'
        ),
    )
    parser.add_argument(
        '--orchestrator-version',
        type=str,
        help=(
            'Specifies the version of the Android Test Orchestrator to use'
            ' during test execution. If not specified, no orchestrator is'
            ' used. If set to `auto`, the system-default orchestrator version'
            ' is used.'
        ),
    )

  def Run(self, args):
    if args.additional_test_options:
      invalid_keys = {
          'package',
          'notPackage',
          'class',
          'notClass',
          'annotation',
          'notAnnotation',
          'size',
      }.intersection(args.additional_test_options.keys())
      if invalid_keys:
        raise calliope_exceptions.InvalidArgumentException(
            '--additional-test-options',
            f'Keys {", ".join(sorted(invalid_keys))} are not allowed. '
            'Use --test-targets instead.',
        )

    if args.sharding_option == 'uniform' and args.uniform_shard_count is None:
      raise calliope_exceptions.RequiredArgumentException(
          '--uniform-shard-count',
          'Required when sharding-option is uniform.',
      )
    if (
        args.sharding_option == 'smart'
        and args.target_smart_shard_time is None
    ):
      raise calliope_exceptions.RequiredArgumentException(
          '--target-smart-shard-time',
          'Required when sharding-option is smart.',
      )
    if (
        args.sharding_option == 'smart'
        and args.smart_sharding_record_name is None
    ):
      raise calliope_exceptions.RequiredArgumentException(
          '--smart-sharding-record-name',
          'Required when sharding-option is smart.',
      )

    location_ref = args.CONCEPTS.location.Parse()
    client = device_run.SessionsClient(api_version='v1alpha')
    messages = client.messages

    storage_client = storage_api.StorageClient()
    bucket_name = args.bucket_name
    if not bucket_name:
      project = properties.VALUES.core.project.Get(required=True)
      bucket_name = f'{project}-devicerun'
      log.status.Print(
          f'Creating default GCS bucket [gs://{bucket_name}] for input and'
          ' result files.'
      )
      if location_ref.locationsId == 'global':
        # Creating STANDARD buckets with locationConstraint GLOBAL is not
        # allowed. Fallback to GCS default location.
        location = None
      else:
        location = location_ref.locationsId
      storage_client.CreateBucketIfNotExists(
          bucket_name,
          project=project,
          location=location,
          check_ownership=True,
      )

    run_id = _GetRunId()
    test_gcs = _UploadFileIfNeeded(
        args.test, bucket_name, storage_client, run_id
    )

    installables = []
    if args.apps:
      for app_apk in args.apps:
        app_apk_gcs = _UploadFileIfNeeded(
            app_apk, bucket_name, storage_client, run_id
        )
        installables.append(
            messages.AndroidInstallable(
                files=[
                    messages.InputFile(
                        gcsInputFile=messages.GcsPath(path=app_apk_gcs)
                    )
                ],
            )
        )

    device_actions = []
    if installables:
      device_action = messages.DeviceAction(
          androidInstallPackages=messages.AndroidInstallPackagesDeviceAction(
              installables=installables
          )
      )
      device_actions.append(device_action)

    if args.other_files_to_push:
      push_action = messages.DeviceAction(
          androidPushFiles=messages.AndroidPushFilesDeviceAction(
              fileConfigs=[
                  messages.FileConfig(
                      destinationPath=dest,
                      sourceFile=messages.InputFile(
                          gcsInputFile=messages.GcsPath(
                              path=_UploadFileIfNeeded(
                                  src, bucket_name, storage_client, run_id
                              )
                          )
                      ),
                  )
                  for src, dest in sorted(args.other_files_to_push.items())
              ]
          )
      )
      device_actions.append(push_action)

    if args.paths_to_pull:
      pull_action = messages.DeviceAction(
          androidPullFiles=messages.AndroidPullFilesDeviceAction(
              paths=args.paths_to_pull
          )
      )
      device_actions.append(pull_action)

    if args.video:
      video_action = messages.DeviceAction(
          androidRecordVideo=messages.AndroidRecordVideoDeviceAction(
              discardOnPass=(args.video == 'on-failure')
          )
      )
      device_actions.append(video_action)

    logcat_action = messages.DeviceAction(
        androidLogcat=messages.AndroidLogcatDeviceAction()
    )
    device_actions.append(logcat_action)

    device_requirement = messages.DeviceRequirement(
        deviceId=args.device
    )
    device_config = messages.DeviceConfig(
        actions=device_actions,
        requirement=device_requirement,
    )
    allocation_config = messages.AllocationConfig(deviceConfigs=[device_config])

    test_installable = messages.AndroidInstallable(
        files=[
            messages.InputFile(gcsInputFile=messages.GcsPath(path=test_gcs))
        ],
    )

    test_targets = []
    if args.test_targets:
      for k, v_or_list in sorted(args.test_targets.items()):
        if isinstance(v_or_list, list):
          for v in v_or_list:
            test_targets.append(f'{k} {v}')
        else:
          test_targets.append(f'{k} {v_or_list}')

    additional_test_options = None
    if args.additional_test_options:
      add_prop_cls = (
          messages.AndroidInstrumentationTest.AdditionalTestOptionsValue
          .AdditionalProperty
      )
      additional_test_options = (
          messages.AndroidInstrumentationTest.AdditionalTestOptionsValue(
              additionalProperties=[
                  add_prop_cls(key=k, value=v)
                  for k, v in sorted(args.additional_test_options.items())
              ]
          )
      )

    test_runner_class = (
        args.test_runner_class if args.test_runner_class else None
    )
    orchestrator_version = (
        args.orchestrator_version if args.orchestrator_version else None
    )
    instrumentation_timeout = (
        f'{args.instrumentation_timeout}s'
        if args.instrumentation_timeout is not None
        else None
    )

    uniform_sharding = None
    smart_sharding = None
    if args.sharding_option == 'uniform':
      uniform_sharding = messages.UniformSharding(
          shardCount=args.uniform_shard_count
      )
    elif args.sharding_option == 'smart':
      record_path = (
          f'gs://{bucket_name}/automation/smart-sharding/'
          f'{args.smart_sharding_record_name}.yaml'
      )
      smart_sharding = messages.SmartSharding(
          targetedShardDuration=f'{args.target_smart_shard_time}s',
          timingRecord=messages.InputFile(
              gcsInputFile=messages.GcsPath(path=record_path)
          ),
      )

    android_instrumentation_test = messages.AndroidInstrumentationTest(
        testInstallable=test_installable,
        testRunnerClass=test_runner_class,
        testTargets=test_targets,
        additionalTestOptions=additional_test_options,
        uniformSharding=uniform_sharding,
        smartSharding=smart_sharding,
        instrumentationTimeout=instrumentation_timeout,
        orchestratorVersion=orchestrator_version,
    )
    job_action = messages.JobAction(
        androidInstrumentationTest=android_instrumentation_test
    )

    labels = None
    if args.labels:
      labels = messages.JobConfig.LabelsValue(
          additionalProperties=[
              messages.JobConfig.LabelsValue.AdditionalProperty(
                  key=k, value=v
              )
              for k, v in sorted(args.labels.items())
          ]
      )

    settings = None
    if args.max_attempts is not None:
      settings = messages.JobSettings(
          retrySettings=messages.RetrySettings(
              flakyTestRetryStrategy=messages.FlakyTestRetryStrategy(
                  flakyTestAttempts=args.max_attempts
              )
          )
      )

    job_config = messages.JobConfig(
        displayName='instrumentation-job',
        allocationConfig=allocation_config,
        action=job_action,
        labels=labels,
        settings=settings,
    )

    # bucket_name and storage_client were initialized earlier in Run

    gcs_path = f'gs://{bucket_name}/automation/sessions'
    output_directory_config = messages.SessionOutputFileDirectoryConfig(
        gcsOutputDirectory=messages.GcsPath(path=gcs_path)
    )
    session_config = messages.SessionConfig(
        displayName='instrumentation-session',
        jobConfigs=[job_config],
        outputDirectoryConfig=output_directory_config,
    )

    session = messages.Session(sessionConfig=session_config)

    session_resp = client.Create(location_ref, session=session)
    log.CreatedResource(session_resp.name, kind='session')
    return session_resp


Instrumentation.detailed_help = {
    'DESCRIPTION':
        'Submit a Device Run session with an instrumentation job.',
    'EXAMPLES':
        """\
To submit an instrumentation session in location `us-central1` on a device
with ID `my-device-id`, run:

  $ {command} --location=us-central1 --device=my-device-id --apps=gs://my-bucket/app.apk --test=gs://my-bucket/test.apk --bucket-name=my-bucket
""",
}



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Resource definitions for Cloud Platform Apis generated from apitools."""

import enum


BASE_URL = 'https://saasservicemgmt.googleapis.com/v1beta1/'
DOCS_URL = 'https://cloud.google.com/saas-runtime/docs'


class Collections(enum.Enum):
  """Collections for all supported apis."""

  PROJECTS = (
      'projects',
      'projects/{projectsId}',
      {},
      ['projectsId'],
      True
  )
  PROJECTS_LOCATIONS = (
      'projects.locations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_FLAGATTRIBUTES = (
      'projects.locations.flagAttributes',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/flagAttributes/'
              '{flagAttributesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_FLAGRELEASES = (
      'projects.locations.flagReleases',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/flagReleases/'
              '{flagReleasesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_FLAGREVISIONS = (
      'projects.locations.flagRevisions',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/flagRevisions/'
              '{flagRevisionsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_FLAGS = (
      'projects.locations.flags',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/flags/{flagsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_RELEASES = (
      'projects.locations.releases',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/releases/'
              '{releasesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_ROLLOUTKINDS = (
      'projects.locations.rolloutKinds',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/rolloutKinds/'
              '{rolloutKindsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_ROLLOUTS = (
      'projects.locations.rollouts',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/rollouts/'
              '{rolloutsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_SAAS = (
      'projects.locations.saas',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/saas/{saasId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_SAASRELEASES = (
      'projects.locations.saasReleases',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/saasReleases/'
              '{saasReleasesId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_TENANTS = (
      'projects.locations.tenants',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/tenants/'
              '{tenantsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_UNITGROUPOPERATIONS = (
      'projects.locations.unitGroupOperations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/'
              'unitGroupOperations/{unitGroupOperationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_UNITGROUPS = (
      'projects.locations.unitGroups',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/unitGroups/'
              '{unitGroupsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_UNITKINDS = (
      'projects.locations.unitKinds',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/unitKinds/'
              '{unitKindsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_UNITOPERATIONS = (
      'projects.locations.unitOperations',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/unitOperations/'
              '{unitOperationsId}',
      },
      ['name'],
      True
  )
  PROJECTS_LOCATIONS_UNITS = (
      'projects.locations.units',
      '{+name}',
      {
          '':
              'projects/{projectsId}/locations/{locationsId}/units/{unitsId}',
      },
      ['name'],
      True
  )

  def __init__(self, collection_name, path, flat_paths, params,
               enable_uri_parsing):
    self.collection_name = collection_name
    self.path = path
    self.flat_paths = flat_paths
    self.params = params
    self.enable_uri_parsing = enable_uri_parsing



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Translation rule for concurrent_requests feature."""

import logging
from typing import Any, Mapping, Sequence

from googlecloudsdk.command_lib.app.gae_to_cr_migration_util.common import util
from googlecloudsdk.command_lib.app.gae_to_cr_migration_util.config import feature_helper


_MAX_CONCURRENT_REQUESTS_KEY = 'automatic_scaling.max_concurrent_requests'
_TARGET_CONCURRENT_REQUESTS_KEY = (
    'automatic_scaling.target_concurrent_requests'
)
_ALLOW_MAX_CONCURRENT_REQ_KEYS = (
    _MAX_CONCURRENT_REQUESTS_KEY,
    _TARGET_CONCURRENT_REQUESTS_KEY,
)
_DEFAULT_STANDARD_CONCURRENCY = 10


def translate_concurrent_requests_features(
    input_data: Mapping[str, any],
    range_limited_features: feature_helper.RangeLimitFeature) -> Sequence[str]:
  """Translates GAE concurrent request settings to Cloud Run.

  This function translates `max_concurrent_requests` (standard) or
  `target_concurrent_requests` (flex) from GAE to the Cloud Run --concurrency
  flag.

  Args:
    input_data: A mapping containing the input configuration.
    range_limited_features: Feature helper for range-limited features.

  Returns:
    A sequence of strings representing the Cloud Run flags.
  """

  is_flex = util.is_flex_env(input_data)

  feature_key = util.get_feature_key_from_input(
      input_data, _ALLOW_MAX_CONCURRENT_REQ_KEYS
  )
  input_has_concurrent_requests = feature_key is not None

  # if input does not have max_concurrent_request/target_concurrent_request
  # specified, use the `automatic_scaling.max_concurrent_requests` from the
  # app2run/config/features.yaml as the default feature.
  if not input_has_concurrent_requests:
    feature = range_limited_features[_MAX_CONCURRENT_REQUESTS_KEY]
    default_value = (
        feature.range['max'] if is_flex else _DEFAULT_STANDARD_CONCURRENCY
    )
    return util.generate_output_flags(feature.flags, default_value)
  feature = range_limited_features[feature_key]
  input_value = input_data[feature_key]
  if input_value < feature.range['min']:
    logging.warning(
        '%s has invalid value of %s, minimum value is %s',
        feature_key, input_value, feature.range['min']
    )
    return []
  if input_value > feature.range['max']:
    logging.warning(
        '%s has invalid value of %s, maximum value is %s.',
        feature_key, input_value, feature.range['max']
    )
    return util.generate_output_flags(feature.flags, feature.range['max'])
  target_value = (
      input_value if feature.validate(input_value) else feature.range['max']
  )
  return util.generate_output_flags(feature.flags, target_value)


def update_service_yaml_with_concurrent_requests(
    service_yaml: dict[str, Any],
    input_data: Mapping[str, Any],
    range_limited_features: Mapping[str, Any],
) -> None:
  """Updates the service_yaml dict with concurrent requests settings."""
  is_flex = util.is_flex_env(input_data)
  feature_key = util.get_feature_key_from_input(
      input_data, _ALLOW_MAX_CONCURRENT_REQ_KEYS
  )
  input_has_concurrent_requests = feature_key is not None

  target_value = None
  if not input_has_concurrent_requests:
    feature = range_limited_features[_MAX_CONCURRENT_REQUESTS_KEY]
    target_value = (
        feature.range['max'] if is_flex else _DEFAULT_STANDARD_CONCURRENCY
    )
  else:
    feature = range_limited_features[feature_key]
    input_value = input_data[feature_key]
    if input_value < feature.range['min']:
      logging.warning(
          '%s has invalid value of %s, minimum value is %s',
          feature_key,
          input_value,
          feature.range['min'],
      )
    elif input_value > feature.range['max']:
      logging.warning(
          '%s has invalid value of %s, maximum value is %s.',
          feature_key,
          input_value,
          feature.range['max'],
      )
      target_value = feature.range['max']
    else:
      target_value = (
          input_value if feature.validate(input_value) else feature.range['max']
      )

  if target_value is not None:
    spec = service_yaml['spec']['template']['spec']
    spec['containerConcurrency'] = int(target_value)



# -*- coding: utf-8 -*- #
# Copyright 2025 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Bigtable materialized-views API helper."""

from typing import Generator
from apitools.base.py import list_pager
from googlecloudsdk.api_lib.bigtable import util
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.util.apis import arg_utils
from googlecloudsdk.core import resources
from googlecloudsdk.generated_clients.apis.bigtableadmin.v2 import bigtableadmin_v2_messages


def Describe(
    materialized_view_ref: resources.Resource, view: str
) -> bigtableadmin_v2_messages.MaterializedView:
  """Describe a materialized view.

  Args:
    materialized_view_ref: A resource reference to the materialized view to
      describe.
    view: Specifies what type of information to return about the view. Valid
      values are 'schema', 'replication', and 'full'.

  Returns:
    materialized view resource object.
  """
  client = util.GetAdminClient()
  msgs = util.GetAdminMessages()

  get_materialized_view_request = (
      msgs.BigtableadminProjectsInstancesMaterializedViewsGetRequest
  )
  view_enum_type = get_materialized_view_request.ViewValueValuesEnum
  mapper = arg_utils.ChoiceEnumMapper(
      '--view',
      view_enum_type,
      custom_mappings={
          'SCHEMA_VIEW': 'schema',
          'REPLICATION_VIEW': 'replication',
          'FULL': 'full',
      },
  )
  view_enum = mapper.GetEnumForChoice(view)
  if view_enum is None:
    view_enum = view_enum_type.SCHEMA_VIEW

  msg = get_materialized_view_request(
      name=materialized_view_ref.RelativeName(), view=view_enum
  )
  return client.projects_instances_materializedViews.Get(msg)


def Create(
    materialized_view_ref: resources.Resource,
    query: str,
    deletion_protection: bool,
    ignore_warnings: bool = False,
) -> bigtableadmin_v2_messages.MaterializedView:
  """Create a materialized view.

  Args:
    materialized_view_ref: A resource reference to the materialized view to
      create.
    query: The query of the materialized view.
    deletion_protection: Whether the materialized view is protected from
      deletion.
    ignore_warnings: Whether to ignore warnings when creating the materialized
      view.

  Returns:
    Created materialized view resource object.
  """

  client = util.GetAdminClient()
  msgs = util.GetAdminMessages()

  instance_ref = materialized_view_ref.Parent()

  materialized_view = msgs.MaterializedView(query=query)
  if deletion_protection is not None:
    materialized_view.deletionProtection = deletion_protection

  msg = msgs.BigtableadminProjectsInstancesMaterializedViewsCreateRequest(
      materializedView=materialized_view,
      materializedViewId=materialized_view_ref.Name(),
      parent=instance_ref.RelativeName(),
  )
  # only set ignore_warnings if it is set to True, since it is behind a
  # visibility label.
  if ignore_warnings:
    msg.ignoreWarnings = ignore_warnings
  return client.projects_instances_materializedViews.Create(msg)


def Delete(
    materialized_view_ref: resources.Resource,
) -> None:
  """Delete a materialized view.

  Args:
    materialized_view_ref: A resource reference to the materialized view to
      delete.
  """
  client = util.GetAdminClient()
  msg = util.GetAdminMessages().BigtableadminProjectsInstancesMaterializedViewsDeleteRequest(
      name=materialized_view_ref.RelativeName()
  )
  client.projects_instances_materializedViews.Delete(msg)


def List(
    instance_ref: resources.Resource,
    view: str,
) -> Generator[bigtableadmin_v2_messages.MaterializedView, None, None]:
  """List materialized views.

  Args:
    instance_ref: A resource reference of the instance to list materialized
      views for.
    view: Specifies what type of information to return about the views. The only
      supported value at this time is 'schema'.

  Returns:
    Generator of materialized view resource objects.
  """
  client = util.GetAdminClient()
  msgs = util.GetAdminMessages()

  list_materialized_view_request = (
      msgs.BigtableadminProjectsInstancesMaterializedViewsListRequest
  )
  view_enum_type = list_materialized_view_request.ViewValueValuesEnum
  mapper = arg_utils.ChoiceEnumMapper(
      '--view',
      view_enum_type,
      custom_mappings={
          'SCHEMA_VIEW': 'schema',
          'REPLICATION_VIEW': 'replication',
          'FULL': 'full',
      },
  )
  view_enum = mapper.GetEnumForChoice(view)
  # Only the schema view is supported for list materialized views at this time.
  if view_enum != view_enum_type.SCHEMA_VIEW:
    raise exceptions.BadArgumentException(
        view,
        "Only 'schema' view is supported for list materialized views.",
    )

  msg = list_materialized_view_request(
      parent=instance_ref.RelativeName(),
      view=view_enum,
  )
  return list_pager.YieldFromList(
      client.projects_instances_materializedViews,
      msg,
      field='materializedViews',
      batch_size_attribute=None,
  )


def Update(
    materialized_view_ref: resources.Resource, deletion_protection: bool
) -> bigtableadmin_v2_messages.MaterializedView:
  """Update a materialized view.

  Args:
    materialized_view_ref: A resource reference to the materialized view to
      update.
    deletion_protection: Whether the materialized view is protected from
      deletion.

  Returns:
    Updated materialized view resource object.
  """

  client = util.GetAdminClient()
  msgs = util.GetAdminMessages()

  msg = msgs.BigtableadminProjectsInstancesMaterializedViewsPatchRequest(
      materializedView=msgs.MaterializedView(
          deletionProtection=deletion_protection
      ),
      name=materialized_view_ref.RelativeName(),
      updateMask='deletion_protection',
  )

  return client.projects_instances_materializedViews.Patch(msg)



# Copyright 2016 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OAuth 2.0 client.

This is a client for interacting with an OAuth 2.0 authorization server's
token endpoint.

For more information about the token endpoint, see
`Section 3.1 of rfc6749`_

.. _Section 3.1 of rfc6749: https://tools.ietf.org/html/rfc6749#section-3.2
"""

import datetime
import http.client as http_client
import json
import logging
import urllib

from google.auth import _exponential_backoff
from google.auth import _helpers
from google.auth import credentials
from google.auth import exceptions
from google.auth import jwt
from google.auth import metrics
from google.auth import transport

_LOGGER = logging.getLogger(__name__)

_URLENCODED_CONTENT_TYPE = "application/x-www-form-urlencoded"
_JSON_CONTENT_TYPE = "application/json"
_JWT_GRANT_TYPE = "urn:ietf:params:oauth:grant-type:jwt-bearer"
_REFRESH_GRANT_TYPE = "refresh_token"
_BLOCKING_REGIONAL_ACCESS_BOUNDARY_LOOKUP_TIMEOUT = 3


def _handle_error_response(response_data, retryable_error):
    """Translates an error response into an exception.

    Args:
        response_data (Mapping | str): The decoded response data.
        retryable_error Optional[bool]: A boolean indicating if an error is retryable.
            Defaults to False.

    Raises:
        google.auth.exceptions.RefreshError: The errors contained in response_data.
    """

    retryable_error = retryable_error if retryable_error else False

    if isinstance(response_data, str):
        raise exceptions.RefreshError(response_data, retryable=retryable_error)
    try:
        error_details = "{}: {}".format(
            response_data["error"], response_data.get("error_description")
        )
    # If no details could be extracted, use the response data.
    except (KeyError, ValueError):
        error_details = json.dumps(response_data)

    raise exceptions.RefreshError(
        error_details, response_data, retryable=retryable_error
    )


def _can_retry(status_code, response_data):
    """Checks if a request can be retried by inspecting the status code
    and response body of the request.

    Args:
        status_code (int): The response status code.
        response_data (Mapping | str): The decoded response data.

    Returns:
      bool: True if the response is retryable. False otherwise.
    """
    if status_code in transport.DEFAULT_RETRYABLE_STATUS_CODES:
        return True

    try:
        # For a failed response, response_body could be a string
        error_desc = response_data.get("error_description") or ""
        error_code = response_data.get("error") or ""

        if not isinstance(error_code, str) or not isinstance(error_desc, str):
            return False

        # Per Oauth 2.0 RFC https://www.rfc-editor.org/rfc/rfc6749.html#section-4.1.2.1
        # This is needed because a redirect will not return a 500 status code.
        retryable_error_descriptions = {
            "internal_failure",
            "server_error",
            "temporarily_unavailable",
        }

        if any(e in retryable_error_descriptions for e in (error_code, error_desc)):
            return True

    except AttributeError:
        pass

    return False


def _parse_expiry(response_data):
    """Parses the expiry field from a response into a datetime.

    Args:
        response_data (Mapping): The JSON-parsed response data.

    Returns:
        Optional[datetime]: The expiration or ``None`` if no expiration was
            specified.
    """
    expires_in = response_data.get("expires_in", None)

    if expires_in is not None:
        # Some services do not respect the OAUTH2.0 RFC and send expires_in as a
        # JSON String.
        if isinstance(expires_in, str):
            expires_in = int(expires_in)

        return _helpers.utcnow() + datetime.timedelta(seconds=expires_in)
    else:
        return None


def _token_endpoint_request_no_throw(
    request,
    token_uri,
    body,
    access_token=None,
    use_json=False,
    can_retry=True,
    headers=None,
    **kwargs
):
    """Makes a request to the OAuth 2.0 authorization server's token endpoint.
    This function doesn't throw on response errors.

    Args:
        request (google.auth.transport.Request): A callable used to make
            HTTP requests.
        token_uri (str): The OAuth 2.0 authorizations server's token endpoint
            URI.
        body (Mapping[str, str]): The parameters to send in the request body.
        access_token (Optional(str)): The access token needed to make the request.
        use_json (Optional(bool)): Use urlencoded format or json format for the
            content type. The default value is False.
        can_retry (bool): Enable or disable request retry behavior.
        headers (Optional[Mapping[str, str]]): The headers for the request.
        kwargs: Additional arguments passed on to the request method. The
            kwargs will be passed to `requests.request` method, see:
            https://docs.python-requests.org/en/latest/api/#requests.request.
            For example, you can use `cert=("cert_pem_path", "key_pem_path")`
            to set up client side SSL certificate, and use
            `verify="ca_bundle_path"` to set up the CA certificates for sever
            side SSL certificate verification.

    Returns:
        Tuple(bool, Mapping[str, str], Optional[bool]): A boolean indicating
          if the request is successful, a mapping for the JSON-decoded response
          data and in the case of an error a boolean indicating if the error
          is retryable.
    """
    if use_json:
        headers_to_use = {"Content-Type": _JSON_CONTENT_TYPE}
        body = json.dumps(body).encode("utf-8")
    else:
        headers_to_use = {"Content-Type": _URLENCODED_CONTENT_TYPE}
        body = urllib.parse.urlencode(body).encode("utf-8")

    if access_token:
        headers_to_use["Authorization"] = "Bearer {}".format(access_token)

    if headers:
        headers_to_use.update(headers)

    response_data = {}
    retryable_error = False

    retries = _exponential_backoff.ExponentialBackoff()
    for _ in retries:
        response = request(
            method="POST", url=token_uri, headers=headers_to_use, body=body, **kwargs
        )
        response_body = (
            response.data.decode("utf-8")
            if hasattr(response.data, "decode")
            else response.data
        )

        try:
            # response_body should be a JSON
            response_data = json.loads(response_body)
        except ValueError:
            response_data = response_body

        if response.status == http_client.OK:
            return True, response_data, None

        retryable_error = _can_retry(
            status_code=response.status, response_data=response_data
        )

        if not can_retry or not retryable_error:
            return False, response_data, retryable_error

    return False, response_data, retryable_error


def _token_endpoint_request(
    request,
    token_uri,
    body,
    access_token=None,
    use_json=False,
    can_retry=True,
    headers=None,
    **kwargs
):
    """Makes a request to the OAuth 2.0 authorization server's token endpoint.

    Args:
        request (google.auth.transport.Request): A callable used to make
            HTTP requests.
        token_uri (str): The OAuth 2.0 authorizations server's token endpoint
            URI.
        body (Mapping[str, str]): The parameters to send in the request body.
        access_token (Optional(str)): The access token needed to make the request.
        use_json (Optional(bool)): Use urlencoded format or json format for the
            content type. The default value is False.
        can_retry (bool): Enable or disable request retry behavior.
        headers (Optional[Mapping[str, str]]): The headers for the request.
        kwargs: Additional arguments passed on to the request method. The
            kwargs will be passed to `requests.request` method, see:
            https://docs.python-requests.org/en/latest/api/#requests.request.
            For example, you can use `cert=("cert_pem_path", "key_pem_path")`
            to set up client side SSL certificate, and use
            `verify="ca_bundle_path"` to set up the CA certificates for sever
            side SSL certificate verification.

    Returns:
        Mapping[str, str]: The JSON-decoded response data.

    Raises:
        google.auth.exceptions.RefreshError: If the token endpoint returned
            an error.
    """

    (
        response_status_ok,
        response_data,
        retryable_error,
    ) = _token_endpoint_request_no_throw(
        request,
        token_uri,
        body,
        access_token=access_token,
        use_json=use_json,
        can_retry=can_retry,
        headers=headers,
        **kwargs
    )
    if not response_status_ok:
        _handle_error_response(response_data, retryable_error)
    return response_data


def jwt_grant(request, token_uri, assertion, can_retry=True):
    """Implements the JWT Profile for OAuth 2.0 Authorization Grants.

    For more details, see `rfc7523 section 4`_.

    Args:
        request (google.auth.transport.Request): A callable used to make
            HTTP requests.
        token_uri (str): The OAuth 2.0 authorizations server's token endpoint
            URI.
        assertion (str): The OAuth 2.0 assertion.
        can_retry (bool): Enable or disable request retry behavior.

    Returns:
        Tuple[str, Optional[datetime], Mapping[str, str]]: The access token,
            expiration, and additional data returned by the token endpoint.

    Raises:
        google.auth.exceptions.RefreshError: If the token endpoint returned
            an error.

    .. _rfc7523 section 4: https://tools.ietf.org/html/rfc7523#section-4
    """
    body = {"assertion": assertion, "grant_type": _JWT_GRANT_TYPE}

    response_data = _token_endpoint_request(
        request,
        token_uri,
        body,
        can_retry=can_retry,
        headers={
            metrics.API_CLIENT_HEADER: metrics.token_request_access_token_sa_assertion()
        },
    )

    try:
        access_token = response_data["access_token"]
    except KeyError as caught_exc:
        new_exc = exceptions.RefreshError(
            "No access token in response.", response_data, retryable=False
        )
        raise new_exc from caught_exc

    expiry = _parse_expiry(response_data)

    return access_token, expiry, response_data


def call_iam_generate_id_token_endpoint(
    request,
    iam_id_token_endpoint,
    signer_email,
    audience,
    access_token,
    universe_domain=credentials.DEFAULT_UNIVERSE_DOMAIN,
):
    """Call iam.generateIdToken endpoint to get ID token.

    Args:
        request (google.auth.transport.Request): A callable used to make
            HTTP requests.
        iam_id_token_endpoint (str): The IAM ID token endpoint to use.
        signer_email (str): The signer email used to form the IAM
            generateIdToken endpoint.
        audience (str): The audience for the ID token.
        access_token (str): The access token used to call the IAM endpoint.
        universe_domain (str): The universe domain for the request. The
            default is ``googleapis.com``.

    Returns:
        Tuple[str, datetime]: The ID token and expiration.
    """
    body = {"audience": audience, "includeEmail": "true", "useEmailAzp": "true"}

    response_data = _token_endpoint_request(
        request,
        iam_id_token_endpoint.replace(
            credentials.DEFAULT_UNIVERSE_DOMAIN, universe_domain
        ).format(signer_email),
        body,
        access_token=access_token,
        use_json=True,
    )

    try:
        id_token = response_data["token"]
    except KeyError as caught_exc:
        new_exc = exceptions.RefreshError(
            "No ID token in response.", response_data, retryable=False
        )
        raise new_exc from caught_exc

    payload = jwt.decode(id_token, verify=False)
    expiry = _helpers.utcfromtimestamp(payload["exp"])

    return id_token, expiry


def id_token_jwt_grant(request, token_uri, assertion, can_retry=True):
    """Implements the JWT Profile for OAuth 2.0 Authorization Grants, but
    requests an OpenID Connect ID Token instead of an access token.

    This is a variant on the standard JWT Profile that is currently unique
    to Google. This was added for the benefit of authenticating to services
    that require ID Tokens instead of access tokens or JWT bearer tokens.

    Args:
        request (google.auth.transport.Request): A callable used to make
            HTTP requests.
        token_uri (str): The OAuth 2.0 authorization server's token endpoint
            URI.
        assertion (str): JWT token signed by a service account. The token's
            payload must include a ``target_audience`` claim.
        can_retry (bool): Enable or disable request retry behavior.

    Returns:
        Tuple[str, Optional[datetime], Mapping[str, str]]:
            The (encoded) Open ID Connect ID Token, expiration, and additional
            data returned by the endpoint.

    Raises:
        google.auth.exceptions.RefreshError: If the token endpoint returned
            an error.
    """
    body = {"assertion": assertion, "grant_type": _JWT_GRANT_TYPE}

    response_data = _token_endpoint_request(
        request,
        token_uri,
        body,
        can_retry=can_retry,
        headers={
            metrics.API_CLIENT_HEADER: metrics.token_request_id_token_sa_assertion()
        },
    )

    try:
        id_token = response_data["id_token"]
    except KeyError as caught_exc:
        new_exc = exceptions.RefreshError(
            "No ID token in response.", response_data, retryable=False
        )
        raise new_exc from caught_exc

    payload = jwt.decode(id_token, verify=False)
    expiry = _helpers.utcfromtimestamp(payload["exp"])

    return id_token, expiry, response_data


def _handle_refresh_grant_response(response_data, refresh_token):
    """Extract tokens from refresh grant response.

    Args:
        response_data (Mapping[str, str]): Refresh grant response data.
        refresh_token (str): Current refresh token.

    Returns:
        Tuple[str, str, Optional[datetime], Mapping[str, str]]: The access token,
            refresh token, expiration, and additional data returned by the token
            endpoint. If response_data doesn't have refresh token, then the current
            refresh token will be returned.

    Raises:
        google.auth.exceptions.RefreshError: If the token endpoint returned
            an error.
    """
    try:
        access_token = response_data["access_token"]
    except KeyError as caught_exc:
        new_exc = exceptions.RefreshError(
            "No access token in response.", response_data, retryable=False
        )
        raise new_exc from caught_exc

    refresh_token = response_data.get("refresh_token", refresh_token)
    expiry = _parse_expiry(response_data)

    return access_token, refresh_token, expiry, response_data


def refresh_grant(
    request,
    token_uri,
    refresh_token,
    client_id,
    client_secret,
    scopes=None,
    rapt_token=None,
    can_retry=True,
):
    """Implements the OAuth 2.0 refresh token grant.

    For more details, see `rfc678 section 6`_.

    Args:
        request (google.auth.transport.Request): A callable used to make
            HTTP requests.
        token_uri (str): The OAuth 2.0 authorizations server's token endpoint
            URI.
        refresh_token (str): The refresh token to use to get a new access
            token.
        client_id (str): The OAuth 2.0 application's client ID.
        client_secret (str): The Oauth 2.0 appliaction's client secret.
        scopes (Optional(Sequence[str])): Scopes to request. If present, all
            scopes must be authorized for the refresh token. Useful if refresh
            token has a wild card scope (e.g.
            'https://www.googleapis.com/auth/any-api').
        rapt_token (Optional(str)): The reauth Proof Token.
        can_retry (bool): Enable or disable request retry behavior.

    Returns:
        Tuple[str, str, Optional[datetime], Mapping[str, str]]: The access
            token, new or current refresh token, expiration, and additional data
            returned by the token endpoint.

    Raises:
        google.auth.exceptions.RefreshError: If the token endpoint returned
            an error.

    .. _rfc6748 section 6: https://tools.ietf.org/html/rfc6749#section-6
    """
    body = {
        "grant_type": _REFRESH_GRANT_TYPE,
        "client_id": client_id,
        "client_secret": client_secret,
        "refresh_token": refresh_token,
    }
    if scopes:
        body["scope"] = " ".join(scopes)
    if rapt_token:
        body["rapt"] = rapt_token

    response_data = _token_endpoint_request(
        request, token_uri, body, can_retry=can_retry
    )
    return _handle_refresh_grant_response(response_data, refresh_token)


def _lookup_regional_access_boundary(request, url, headers=None, fail_fast=False):
    """Implements the global lookup of a credential Regional Access Boundary.
    For the lookup, we send a request to the global lookup endpoint and then
    parse the response. Service account credentials, workload identity
    pools and workforce pools implementation may have Regional Access Boundaries configured.
    Args:
        request (google.auth.transport.Request): A callable used to make
            HTTP requests.
        url (str): The Regional Access Boundary lookup url.
        headers (Optional[Mapping[str, str]]): The headers for the request.
        fail_fast (bool): Whether the lookup should fail fast (uses a short timeout and no retries).
    Returns:
        Optional[Mapping[str,list|str]]: A dictionary containing
            "locations" as a list of allowed locations as strings and
            "encodedLocations" as a hex string.
            e.g:
            {
                "locations": [
                    "us-central1", "us-east1", "europe-west1", "asia-east1"
                ],
                "encodedLocations": "0xA30"
            }
    """

    response_data = _lookup_regional_access_boundary_request(
        request, url, headers=headers, fail_fast=fail_fast
    )
    if response_data is None:
        # Error was already logged by _lookup_regional_access_boundary_request
        return None

    if not isinstance(response_data, dict) or "encodedLocations" not in response_data:
        _LOGGER.error(
            "Regional Access Boundary response malformed: missing 'encodedLocations' key in %s",
            response_data,
        )
        return None
    return response_data


def _lookup_regional_access_boundary_request(
    request, url, can_retry=True, headers=None, fail_fast=False
):
    """Makes a request to the Regional Access Boundary lookup endpoint.

    Args:
        request (google.auth.transport.Request): A callable used to make
            HTTP requests.
        url (str): The Regional Access Boundary lookup url.
        can_retry (bool): Enable or disable request retry behavior. Defaults to true.
        headers (Optional[Mapping[str, str]]): The headers for the request.
        fail_fast (bool): Whether the lookup should fail fast (uses a short timeout and no retries).

    Returns:
        Optional[Mapping[str, str]]: The JSON-decoded response data on success, or None on failure.
    """
    (
        response_status_ok,
        response_data,
        retryable_error,
    ) = _lookup_regional_access_boundary_request_no_throw(
        request, url, can_retry=can_retry, headers=headers, fail_fast=fail_fast
    )
    if not response_status_ok:
        _LOGGER.warning(
            "Regional Access Boundary HTTP request failed after retries: response_data=%s, retryable_error=%s",
            response_data,
            retryable_error,
        )
        return None
    return response_data


def _lookup_regional_access_boundary_request_no_throw(
    request, url, can_retry=True, headers=None, fail_fast=False
):
    """Makes a request to the Regional Access Boundary lookup endpoint. This
        function doesn't throw on response errors.

    Args:
        request (google.auth.transport.Request): A callable used to make
            HTTP requests.
        url (str): The Regional Access Boundary lookup url.
        can_retry (bool): Enable or disable request retry behavior. Defaults to true.
        headers (Optional[Mapping[str, str]]): The headers for the request.
        fail_fast (bool): Whether the lookup should fail fast (uses a short timeout and no retries).

    Returns:
        Tuple(bool, Mapping[str, str], Optional[bool]): A boolean indicating
          if the request is successful, a mapping for the JSON-decoded response
          data and in the case of an error a boolean indicating if the error
          is retryable.
    """

    response_data = {}
    retryable_error = False

    timeout = _BLOCKING_REGIONAL_ACCESS_BOUNDARY_LOOKUP_TIMEOUT if fail_fast else None
    total_attempts = 1 if fail_fast else 6
    retries = _exponential_backoff.ExponentialBackoff(total_attempts=total_attempts)

    for _ in retries:
        response = request(method="GET", url=url, headers=headers, timeout=timeout)
        response_body = (
            response.data.decode("utf-8")
            if hasattr(response.data, "decode")
            else response.data
        )

        try:
            # response_body should be a JSON
            response_data = json.loads(response_body)
        except ValueError:
            response_data = response_body

        if response.status == http_client.OK:
            return True, response_data, None

        retryable_error = _can_retry(
            status_code=response.status, response_data=response_data
        )
        # Add 502 (Bad Gateway) as a retryable error for RAB lookups.
        if response.status == http_client.BAD_GATEWAY:
            retryable_error = True

        if not can_retry or not retryable_error:
            return False, response_data, retryable_error

    return False, response_data, retryable_error
